# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
# pylint: disable=import-outside-toplevel

from dataclasses import dataclass
from typing import Iterable, Union
import colemen_utils as c
import config
import silent.Package as _package


@dataclass
class Main:
    project_name:str = None
    root_path:str = None

    def __init__(self):
        self._entities = {
            "packages":{},
            "modules":{},
            "imports":[],
            "classes":[],
            "methods":[],
        }
        self._packages = {}
        self.data = {}


    def master(self):
        print("master")

    @property
    def summary(self):
        '''
            Get this main's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 09:39:11
            `@memberOf`: main
            `@property`: summary
        '''
        value = {
            "packages":{}
        }

        for k,v in self._packages.items():
            value['packages'][k] = v.summary
        c.file.writer.to_json("tmp.json",value)
        return value

    def save(self):
        imports = []
        for pkg in self.packages:
            pkg.save()
            imports.append(pkg.import_statement)
        imports = '\n'.join(imports)
        c.file.write(f"{self.root_path}/{self.project_name}/__init__.py",imports)


    def add_package(self,name:str,description:str=None)->config._package_type:
        '''
            add a new package to this project.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the package

            [`description`=None] {str}
                The documentation description for the package.


            Return {Package}
            ----------------------
            The new package instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-05-2023 09:09:47
            `memberOf`: main
            `version`: 1.0
            `method_name`: add_package
            * @xxx [01-05-2023 09:11:21]: documentation for add_package
        '''
        file_path = f"{self.root_path}/{self.project_name}/{name}"
        pkg = _package.Package(
            self,
            name=name,
            description=description,
            file_path=file_path,
        )
        self._packages[name] = pkg
        return pkg


    def get_modules_by_tag(self,tag:str)->Iterable[config._py_module_type]:
        '''
            Get this main's get_modules_by_tag

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 09:59:00
            `@memberOf`: main
            `@property`: get_modules_by_tag
        '''
        value = []
        for mod in self.modules:
            if mod.has_tag(tag):
                value.append(mod)
        return value

    def get_methods_by_tag(self,tag:Union[str,list],match_all:bool=False)->Iterable[config._method_type]:
        '''
            Retrieve all methods with matching tags
            ----------

            Arguments
            -------------------------
            `tag` {str,list}
                The tag or list of tags to search for.

            [`match_all`=False] {bool}
                If True, all tags provided must be found.

            Return {list}
            ----------------------
            A list of methods that contain the matching tags, an empty list if None are found.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-06-2023 10:02:34
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: get_methods_by_tag
            * @xxx [01-06-2023 10:04:42]: documentation for get_methods_by_tag
        '''

        value = []
        for mod in self.methods:
            if mod.has_tag(tag):
                value.append(mod,match_all)
        return value

    def get_by_tag(self,tag,entity_type:str=None):
        results = []
        entity_type = c.string.to_snake_case(entity_type)
        tags = c.arr.force_list(tag)
        tags.sort()
        tags.sort(key=len, reverse=False)

        for pkg in self.packages:
            if pkg.has_tag(tags):
                results.append(pkg)

        return results


    @property
    def packages(self)->Iterable[config._package_type]:
        '''
            Get a list of this project's packages

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 09:48:14
            `@memberOf`: main
            `@property`: packages
        '''
        # value = self._entities['packages']
        value = list(self._entities['packages'].values())
        return value

    @property
    def modules(self)->Iterable[config._py_module_type]:
        '''
            Get a list of this project's modules

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 09:48:14
            `@memberOf`: main
            `@property`: modules
        '''
        value = list(self._entities['modules'].values())
        return value

    @property
    def classes(self)->Iterable[config._py_class_type]:
        '''
            Get a list of this project's classes

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 09:48:14
            `@memberOf`: main
            `@property`: classes
        '''
        value = self._entities['classes']
        return value


    @property
    def methods(self)->Iterable[config._method_type]:
        '''
            Get a list of this project's classes

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 09:48:14
            `@memberOf`: main
            `@property`: classes
        '''
        value = self._entities['methods']
        return value

    def register(self,instance):
        from silent.Package import Package
        if isinstance(instance,Package):
            self._entities['packages'][instance.name.name] = instance

        from silent.Module import Module
        if isinstance(instance,Module):
            self._entities['modules'][instance.name.name] = instance

        from silent.Import import ImportStatement
        if isinstance(instance,ImportStatement):
            self._entities['imports'].append(instance)

        from silent.Class import Class
        if isinstance(instance,Class):
            self._entities['classes'].append(instance)

        from silent.Method.Method import Method
        if isinstance(instance,Method):
            self._entities['methods'].append(instance)


def new(project_name:str,root_path:str):
    m = Main()
    m.project_name = project_name
    m.root_path = root_path
    return m



# if __name__ == '__main__':
#     m = Main()
#     m.master()




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import



import ast
from dataclasses import dataclass
from string import Template
from typing import Iterable, Union
import silent.EntityBase as _eb

import colemen_utils as c

# from silent.EntityBase import EntityBase as _eb
# import agod.TypeDeclaration as _type_dec
# from config import column_type,_relationship_type,endpointdoc_type,route_type,root_directory,susurrus_type
import config
log = c.con.log

_STANDARD_LIBRARY_NAMES =["string","re","difflib","textwrap","unicodedata","stringprep","readline","rlcompleter","struct","codecs","datetime","zoneinfo","calendar","collections","collections.abc","heapq","bisect","array","weakref","types","copy","pprint","reprlib","enum","graphlib","numbers","math","cmath","decimal","fractions","random","statistics","itertools","functools","operator","pathlib","os.path","fileinput","stat","filecmp","tempfile","glob","fnmatch","linecache","shutil","pickle","copyreg","shelve","marshal","dbm","sqlite3","zlib","gzip","bz2","lzma","zipfile","tarfile","csv","configparser","tomllib","netrc","plistlib","hashlib","hmac","secrets","os","io","time","argparse","getopt","logging","logging.config","logging.handlers","getpass","curses","curses.textpad","curses.ascii","curses.panel","platform","errno","ctypes","threading","multiprocessing","multiprocessing.shared_memory","concurrent.futures","subprocess","sched","queue","contextvars","_thread","asyncio","socket","ssl","select","selectors","signal","mmap","email","json","mailbox","mimetypes","base64","binascii","quopri","html","html.parser","html.entities","xml.etree.ElementTree","xml.dom","xml.dom.minidom","xml.dom.pulldom","xml.sax","xml.sax.handler","xml.sax.saxutils","xml.sax.xmlreader","xml.parsers.expat","webbrowser","wsgiref","urllib","urllib.request","urllib.response","urllib.parse","urllib.error","urllib.robotparser","http","http.client","ftplib","poplib","imaplib","smtplib","uuid","socketserver","http.server","http.cookies","http.cookiejar","xmlrpc","xmlrpc.client","xmlrpc.server","ipaddress","wave","colorsys","gettext","locale","turtle","cmd","shlex","tkinter","tkinter.colorchooser","tkinter.font","tkinter.messagebox","tkinter.scrolledtext","tkinter.dnd","tkinter.ttk","tkinter.tix","typing","pydoc","doctest","unittest","unittest.mock","unittest.mock","2to3","test","test.support","test.support.socket_helper","test.support.script_helper","test.support.bytecode_helper","test.support.threading_helper","test.support.os_helper","test.support.import_helper","test.support.warnings_helper","bdb","faulthandler","pdb","timeit","trace","tracemalloc","distutils","ensurepip","venv","zipapp","sys","sysconfig","builtins","__main__","warnings","dataclasses","contextlib","abc","atexit","traceback","__future__","gc","inspect","site","code","codeop","zipimport","pkgutil","modulefinder","runpy","importlib","ast","symtable","token","keyword","tokenize","tabnanny","pyclbr","py_compile","compileall","dis","pickletools","msvcrt","winreg","winsound","posix","pwd","grp","termios","tty","pty","fcntl","resource","syslog","aifc","asynchat","asyncore","audioop","cgi","cgitb","chunk","crypt","imghdr","imp","mailcap","msilib","nis","nntplib","optparse","ossaudiodev","pipes","smtpd","sndhdr","spwd","sunau","telnetlib","uu","xdrlib"]

@dataclass
class ImportStatement(_eb.EntityBase):

    import_path:str = None
    '''The import path'''

    alias:str = None
    '''The alias to use for this import'''

    _subjects:Iterable[str] = None
    '''The named subjects to be imported'''

    is_standard_library:bool = False
    '''True if this is importing from the python standard library'''

    is_third_party:bool = False
    '''True if this is importing from a third party library'''



    def __init__(self,**kwargs) -> None:
        '''
            Represents an import statement.

            ----------

            Keyword Arguments
            ----------
            [`import_path`=None] {str}
                The import path where the imported value is located.

            [`alias`=None] {str}
                The alias to use for this import

            [`subjects`=None] {list,str}
                The subject or list of subjects to import

            [`is_standard_library`=False] {bool}
                True if this import is from the python standard library.

            [`is_third_party`=False] {bool}
                True if this is importing from a third party library.


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-26-2022 08:35:16
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: Table
            * @xxx [12-26-2022 08:36:08]: documentation for Table
        '''
        super().__init__(**kwargs)
        self._subjects = []









    @property
    def summary(self):
        '''
            Get the summary property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-06-2022 12:10:00
            `@memberOf`: __init__
            `@property`: summary
        '''
        value = {
            "subjects":self.subjects,
            "import_path":self.import_path,
            "alias":self.alias,
            "is_standard_library":self.is_standard_library,
            "is_third_party":self.is_third_party,
            # "schema":self.table.database.database,
        }

        return value

    def add_subject(self,name):
        '''
            Add a subject to import.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 10:11:34
            `@memberOf`: ImportStatement
            `@property`: add_subject
        '''
        self._subjects.append(name)
        self._subjects = c.arr.remove_duplicates(self.subjects)

    @property
    def subjects(self)->Iterable[str]:
        '''
            Get this ImportStatement's subjects

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 10:52:48
            `@memberOf`: ImportStatement
            `@property`: subjects
        '''
        value = list(self._subjects)
        return value

    @subjects.setter
    def subjects(self,value:Union[str,list]):
        '''
            Set the ImportStatement's subjects property

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 10:55:27
            `@memberOf`: ImportStatement
            `@property`: subjects
        '''
        self._subjects = []
        if isinstance(value,(str)):
            value = value.split(",")
        value = c.arr.force_list(value)
        self._subjects = value

    # @property
    # def _subject_names(self):
    #     '''
    #         Get this ImportStatement's _subject_names

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-04-2023 10:08:17
    #         `@memberOf`: ImportStatement
    #         `@property`: _subject_names
    #     '''
    #     subjects = c.arr.force_list(self.subjects)
    #     value = ""
    #     if len(subjects) > 0:
    #         value = ','.join(subjects)
    #     return value

    # @property
    # def _alias(self):
    #     '''
    #         Get this ImportStatement's _alias

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-04-2023 10:10:10
    #         `@memberOf`: ImportStatement
    #         `@property`: _alias
    #     '''
    #     value = ""
    #     if len(self.subjects) == 1:
    #         value = f"as {self.alias}"
    #     return value

    def __gen_ast_names(self):
        output = []
        for name in self.subjects:
            alias = ast.alias(name=name)
            if self.alias is not None:
                alias.asname=self.alias
            output.append(alias)
            # print(ast.dump(alias,indent=4))
        return output

    @property
    def ast(self)->Union[ast.Import,ast.ImportFrom]:
        '''
            Get this ImportStatement's ast object.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 10:57:49
            `@memberOf`: ImportStatement
            `@property`: ast
        '''
        imp = None
        if self.import_path is not None:
            imp = ast.ImportFrom()
            imp.module=self.import_path
            imp.level=0
            imp.names = self.__gen_ast_names()
        else:
            imp = ast.Import()
            imp.names = self.__gen_ast_names()
        imp = ast.fix_missing_locations(imp)
        return imp

    @property
    def result(self):
        '''
            Get the python source code of this import statement.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 09:17:47
            `@memberOf`: Property
            `@property`: declaration
        '''
        # imp = None
        # if self.import_path is not None:
        #     imp = ast.ImportFrom()
        #     imp.module=self.import_path
        #     imp.level=0
        #     imp.names = self.__gen_ast_names()
        # else:
        #     imp = ast.Import()
        #     imp.names = self.__gen_ast_names()
        # imp = ast.fix_missing_locations(imp)
        # # print(ast.dump(imp,indent=4))
        return ast.unparse(self.ast)




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import




import datetime
import ast
from dataclasses import dataclass
from string import Template
from typing import Iterable, Union

import colemen_utils as c
import silent.EntityBase as _eb


import silent.Class.Property as _prop
import silent.Method.Method as _method

import config
log = c.con.log


@dataclass
class Class(_eb.EntityBase):

    # module:config._py_module_type = None
    # '''The module that this class belongs to'''

    _properties:Iterable[config._py_property_type] = None
    '''A dictionary of property instances associated to this class.'''


    _methods:Iterable[config._method_type] = None
    '''A dictionary of methods that belong to this class.'''

    bases:Iterable[str] = None
    '''A list of classes that this class bases'''

    is_dataclass:bool = False
    '''True if this class is a dataclass'''

    # description:str = None
    # '''The description for this class's docblock'''

    init_body:str = None
    '''The body to apply to the init method.'''

    # _args = None
    # _kwargs = None

    def __init__(self,main:config._master_type,module:config._py_module_type,**kwargs) -> None:
        '''
            Represents a python class

            Arguments
            ------------------
            `main` {Main}
                The project that this class belongs to.

            `module` {Module}
                The module instance that this class belongs to.


            Keyword Arguments
            ------------------
            [`name`=None] {str}
                The name of this class

            [`description`=None] {str}
                The docblock description of this class

            [`init_body`=None] {str}
                The body of the __init__ method.

            [`is_dataclass`=False] {bool}
                True if this class is a dataclass

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-26-2022 08:35:16
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: Table
            * @xxx [12-26-2022 08:36:08]: documentation for Table
        '''
        kwargs['main'] = main
        kwargs['module'] = module
        super().__init__(**kwargs)

        # self.main:config._master_type = main

        self.bases = []
        self._properties = {}
        self._methods = {}
        
        if self.is_dataclass is True:
            self.module.add_import("dataclasses","dataclass")


    @property
    def summary(self):
        '''
            Get the summary property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-06-2022 12:10:00
            `@memberOf`: __init__
            `@property`: summary
        '''
        value = {
            "name":self.name.name,
            "description":self.description,
            "is_dataclass":self.is_dataclass,
            "properties":[],
            "methods":[],
        }
        
        for prop in self.properties:
            value['properties'].append(prop.summary)
        for method in self.methods:
            value['methods'].append(method.summary)
            

        return value

    def add_base(self,class_name:str):
        self.bases.append(class_name)



    # ---------------------------------------------------------------------------- #
    #                                  PROPERTIES                                  #
    # ---------------------------------------------------------------------------- #

    def add_property(self,name:str,**kwargs)->config._py_property_type:
        '''
            Add a property to this class
            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the property to add.

            Keyword Arguments
            -------------------------
            [`data_type`=None] {str}
                The python data type stored in this property

            [`default`=None] {any}
                The default value to assign to this property

            [`description`=None] {str}
                The Docblock description for this property

            [`private`=False] {bool}
                True if this property is private to its class.

            [`getter_body`=None] {str}
                The body of the getter method the default will return the property.

            Return {Property}
            ----------------------
            The newly instantiated property.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-04-2023 09:23:52
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_property
            * @xxx [01-04-2023 09:26:32]: documentation for add_property
        '''
        prop = _prop.Property(self.main,self,
            name = name,
            data_type = c.obj.get_kwarg(['data_type','type'],None,(str),**kwargs),
            default = c.obj.get_kwarg(['default','default_value'],None,None,**kwargs),
            description = c.obj.get_kwarg(['description'],None,(str),**kwargs),
            is_private = c.obj.get_kwarg(['is_private','private'],False,(bool),**kwargs),
            getter_body = c.obj.get_kwarg(['getter_body'],None,(str),**kwargs),
        )


        self._properties[name] = prop

        return prop



    @property
    def properties(self)->Iterable[config._py_property_type]:
        '''
            Get a list of properties for this class.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 09:30:42
            `@memberOf`: __init__
            `@property`: properties
        '''
        value = list(self._properties.values())
        return value

    @property
    def property_names(self)->Iterable[str]:
        '''
            Get the property_names property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 07:37:35
            `@memberOf`: __init__
            `@property`: property_names
        '''
        value = [x.name.name for x in self.properties]
        return value

    @property
    def private_props(self)->Iterable[config._py_property_type]:
        '''
            Get the private properties for this class

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 09:28:10
            `@memberOf`: __init__
            `@property`: private_props
        '''
        value = []
        for prop in self.properties:
            if prop.is_private:
                value.append(prop)
        return value

    @property
    def _dataclass_properties(self)->Iterable[config._py_property_type]:
        '''
            Get a list of properties that are handled by the dataclass.

            These properties are assigned prior to the init method so that the dataclass will automatically handle
            the getters and setters.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 11:04:35
            `@memberOf`: __init__
            `@property`: _dataclass_properties
        '''
        if self.is_dataclass is False:
            return None
        value = []
        for prop in self.properties:
            value.append(prop)
        return value

    def get_property(self,name:str):
        '''
            Retrieve a property from this class
            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the property to search for.

            Return {Property}
            ----------------------
            The property instance, if it exists, None otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-04-2023 09:02:11
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: get_property
            * @xxx [01-04-2023 09:02:58]: documentation for get_property
        '''
        if name in self._properties:
            return self._properties[name]
        if f"_{name}" in self._properties:
            return self._properties[f"_{name}"]



    # ---------------------------------------------------------------------------- #
    #                                    METHODS                                   #
    # ---------------------------------------------------------------------------- #

    def add_method(self,name,**kwargs)->config._method_type:
        '''
            Add a method to this class.
            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the method.

            Keyword Arguments
            -------------------------
            [`description`=None] {str}
                The docblock description for this method.

            [`body`="pass"] {str}
                The body of the method.

            [`return_type`=None] {str}
                The type returned by this method

            [`is_getter`=False] {bool}
                If True the property decorator is added to the method

            [`is_setter`=False] {bool}
                If True the property setter decorator is added to the method

            Return {Method}
            ----------------------
            The new method instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-04-2023 13:28:59
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_method
            * @TODO []: documentation for add_method
        '''
        # kwargs['name'] = name
        method = _method.Method(
            self.main,
            pyclass=self,
            package=self.package,
            name=name,
            description=c.obj.get_kwarg(['description'],None,(str),**kwargs),
            body=c.obj.get_kwarg(['body'],None,(str),**kwargs),
            return_type=c.obj.get_kwarg(['return_type'],None,(str),**kwargs),
            is_getter=c.obj.get_kwarg(['is_getter'],False,(bool),**kwargs),
            is_setter=c.obj.get_kwarg(['is_setter'],False,(bool),**kwargs),
            is_class_method=True,
        )

        self._methods[method.local_name] = method
        return method

    def add_getter(self,name,**kwargs)->config._method_type:
        '''
            Add a getter method to this class.
            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the method.

            Keyword Arguments
            -------------------------
            [`description`=None] {str}
                The docblock description for this method.

            [`body`="pass"] {str}
                The body of the method.

            [`return_type`=None] {str}
                The type returned by this method

            Return {Method}
            ----------------------
            The new method instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-04-2023 13:28:59
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_method
            * @TODO []: documentation for add_method
        '''
        # kwargs['name'] = name
        method = _method.Method(
            self.main,
            pyclass=self,
            package=self.package,
            name=name,
            description=c.obj.get_kwarg(['description'],None,(str),**kwargs),
            body=c.obj.get_kwarg(['body'],None,(str),**kwargs),
            return_type=c.obj.get_kwarg(['return_type'],None,(str),**kwargs),
            is_getter=True,
            is_setter=False,
            is_class_method=True,
        )
        self._methods[method.local_name] = method
        return method

    def add_setter(self,name,**kwargs)->config._method_type:
        '''
            Add a setter method to this class.
            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the method.

            Keyword Arguments
            -------------------------
            [`description`=None] {str}
                The docblock description for this method.

            [`body`="pass"] {str}
                The body of the method.

            [`return_type`=None] {str}
                The type returned by this method

            Return {Method}
            ----------------------
            The new method instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-04-2023 13:28:59
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_method
            * @TODO []: documentation for add_getter
        '''
        # kwargs['name'] = name
        method = _method.Method(
            self.main,
            pyclass=self,
            package=self.package,
            name=name,
            description=c.obj.get_kwarg(['description'],None,(str),**kwargs),
            body=c.obj.get_kwarg(['body'],None,(str),**kwargs),
            return_type=c.obj.get_kwarg(['return_type'],None,(str),**kwargs),
            is_getter=False,
            is_setter=True,
            is_class_method=True,
        )
        self._methods[method.local_name] = method
        return method


    def get_method(self,name:str,partial_match=False)->config._method_type:
        '''
            retrieve a method by its name

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the method to search for.

            [`partial_match`=True] {bool}
                If True, the method name must contain the searched name


            Return {Method}
            ----------------------
            The method instance, None if it is not found.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-06-2023 08:55:24
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: get_method
            * @xxx [01-06-2023 08:56:24]: documentation for get_method
        '''
        if partial_match is False:
            if name in self._methods:
                return self._methods[name]

            get_loc_name = f"{name}____get"
            if get_loc_name in self._methods:
                return self._methods[get_loc_name]

            set_loc_name = f"{name}____set"
            if set_loc_name in self._methods:
                return self._methods[set_loc_name]
        else:
            for k,v in self._methods.items():
                if name in k:
                    return v

    def get_methods_by_tag(self,tag:Union[str,list],match_all:bool=False)->Iterable[config._method_type]:
        '''
            Retrieve all methods with matching tags
            ----------

            Arguments
            -------------------------
            `tag` {str,list}
                The tag or list of tags to search for.

            [`match_all`=False] {bool}
                If True, all tags provided must be found.

            Return {list}
            ----------------------
            A list of methods that contain the matching tags, an empty list if None are found.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-06-2023 10:02:34
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: get_methods_by_tag
            * @xxx [01-06-2023 10:04:42]: documentation for get_methods_by_tag
        '''
        value = []
        for mod in self.methods:
            if mod.has_tag(tag,match_all):
                value.append(mod)
        return value

    @property
    def method_names(self)->Iterable[str]:
        '''
            Get a list of method names associated to this class.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 07:38:46
            `@memberOf`: __init__
            `@property`: method_names
        '''

        value = [x.name for x in self.methods]
        return value

    @property
    def methods(self)->Iterable[config._method_type]:
        '''
            Get a list of method instances associated to this class.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 13:37:45
            `@memberOf`: __init__
            `@property`: methods
        '''
        value = list(self._methods.values())
        return value



    @property
    def _gen_getter_methods(self):
        '''
            Get the _gen_getter_methods property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 13:33:41
            `@memberOf`: __init__
            `@property`: _gen_getter_methods
        '''
        value = []
        for method in self.methods:
            if method.is_getter:
                value.append(ast.unparse(method.declaration_ast))
        return '\n'.join(value)

    @property
    def _gen_setter_methods(self):
        '''
            Get the _gen_setter_methods property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 13:33:41
            `@memberOf`: __init__
            `@property`: _gen_setter_methods
        '''
        value = []
        for method in self.methods:
            if method.is_setter:
                value.append(ast.unparse(method.declaration_ast))
        return '\n'.join(value)

    @property
    def _gen_misc_methods(self):
        '''
            Get the _gen_setter_methods property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 13:33:41
            `@memberOf`: __init__
            `@property`: _gen_setter_methods
        '''
        value = []
        for method in self.methods:
            if method.name.name == "__init__":
                continue
            if method.is_setter is False and method.is_getter is False:
                value.append(ast.unparse(method.declaration_ast))
        return '\n'.join(value)




    @property
    def gen_init(self)->config._method_type:
        '''
            Get the __init__ method for this class or create it.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 14:29:44
            `@memberOf`: __init__
            `@property`: gen_init
        '''

        value = self.get_method("__init__")
        if value is None:
            initsuper = "super().__init__()" if len(self.bases) else ""

            ib = self.init_body
            if ib is None:
                ib = "pass"
            if len(initsuper) > 0:
                if "super().__init__" not in ib:
                    ib = f"{initsuper}\n{ib}"
                    # @Mstep [IF] if the last line of the body is "pass"
                    if ib.split("\n")[-1] == "pass":
                        # @Mstep [] remove the last line from the body.
                        ib = '\n'.join(ib.split("\n")[:-1])

            value = self.add_method(
                "__init__",
                description=self.description,
                body=ib
            )
        return value







    # @property
    # def _gen_bases(self):
    #     '''
    #         Generate a comma delimited list of classes that are extended by this class.

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-04-2023 09:37:05
    #         `@memberOf`: __init__
    #         `@property`: _gen_bases
    #     '''
    #     value = ""
    #     if len(self.bases) > 0:
    #         el = ','.join(self.bases)
    #         value = f"({el})"
    #     return value

    # ---------------------------------------------------------------------------- #
    #                                   __INIT__                                   #
    # ---------------------------------------------------------------------------- #
    
    

    def add_arg(self,name,**kwargs):
        '''
            Add an argument to this class's init method.

            If the name matches a property it will be automatically assigned in the __init__ method.
            
            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the argument.

            Keyword Arguments
            -------------------------
            [`data_type`=None] {str}
                The expected type of the argument

            [`default_value`=None] {any}
                The default value to use if no value is provided.

            [`description`=None] {any}
                The documentation description

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-04-2023 14:46:25
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_arg
            * @xxx [01-05-2023 07:47:58]: documentation for add_arg
        '''
        value = self.gen_init
        value.add_arg(name,**kwargs)

    @property
    def arguments(self)->Iterable[config._method_argument_type]:
        '''
            Get the arguments property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 11:31:10
            `@memberOf`: __init__
            `@property`: arguments
        '''
        value = self.gen_init.arg_list
        return value

    def add_kwarg(self,name,**kwargs):
        '''
            Add a keyword argument to this class's init method.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the argument.


            Keyword Arguments
            -------------------------
            [`data_type`=None] {str}
                The expected type of the argument

            [`default_value`=None] {any}
                The default value to use if no value is provided.

            [`description`=None] {any}
                The documentation description


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-04-2023 14:45:43
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_kwarg
            * @TODO []: documentation for add_kwarg
        '''
        value = self.gen_init
        value.add_kwarg(name,**kwargs)

    def _apply_init_arg_assignments(self):
        '''
            Add property assignments when the init method has an argument with name matching a class property.

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-05-2023 07:46:39
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: _apply_init_arg_assignments
            * @xxx [01-05-2023 07:47:43]: documentation for _apply_init_arg_assignments
        '''
        init = self.gen_init
        for arg in init.arg_list:
            for prop in self.properties:
                if arg.name.name == prop.name.name:
                    arg = init.get_arg(prop.name.name)
                    if arg is None:
                        continue
                    arg.data_type = prop.data_type
                    arg.description = prop.description
                    arg.default = prop.default
                    # print(f"init.body:{init.body}")
                    assign = prop.assign_result(arg.name.name,True)
                    # assign = f"self.{prop.attribute_name}{prop.attribute_type} = {arg.name.name}"
                    if assign not in init.body:
                        # @Mstep [IF] if the last line of the body is "pass"
                        if init.body.split("\n")[-1] == "pass":
                            # @Mstep [] remove the last line from the body.
                            init.body = '\n'.join(init.body.split("\n")[:-1])
                        # @Mstep [] concatenate the property assignment with the init body.
                        init.body=f"{init.body}\n{assign}"

                # if self.is_dataclass is False:
                #     assign = prop.assign_result(arg.name.name,True)
    @property
    def _gen_ast_bases(self):
        '''
            Get the _gen_ast_bases property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 11:57:15
            `@memberOf`: __init__
            `@property`: _gen_ast_bases
        '''
        bases = []
        for base in self.bases:
            bases.append(ast.Name(id=base,ctx=ast.Load()))
        return bases


    @property
    def ast(self):
        '''
            Get this class's ast object.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 11:40:02
            `@memberOf`: __init__
            `@property`: ast
        '''
        value = ast.ClassDef(
            name=self.name.name,
            bases=self._gen_ast_bases,
            keywords=[],
            body=[],
            decorator_list=[],
        )

        if self.is_dataclass is True:
            value.decorator_list.append(ast.Name(id='dataclass',ctx=ast.Load()))

        # xxx [01-05-2023 15:06:43]: apply properties to body
        if self.is_dataclass:
            for prop in self.properties:
                past = prop.declaration_ast
                if isinstance(past,(list)):
                    value.body = value.body + past
                else:
                    value.body.append(past)
        else:
            init = self.gen_init
            for prop in self.properties:
                if init.get_arg(prop.name.name) is not None:
                    continue

                assign = prop.assign_result(prop.default,True)
                if assign not in init.body:
                    # @Mstep [IF] if the last line of the body is "pass"
                    if init.body.split("\n")[-1] == "pass":
                        # @Mstep [] remove the last line from the body.
                        init.body = '\n'.join(init.body.split("\n")[:-1])
                    # @Mstep [] concatenate the property assignment with the init body.                    
                    init.body = f"{init.body}\n{assign}"

        self._apply_init_arg_assignments()
        # self.gen_init.Doc.description = self.description
        # self.gen_init.Doc.subject_name = self.name.name
        # xxx [01-05-2023 15:06:38]: apply init to body.
        value.body.append(self.gen_init.declaration_ast)

        # TODO []: apply methods to body
        # @Mstep [] add the custom getter methods
        value.body.append(ast.parse(self._gen_getter_methods))
        
        # @Mstep [] add the custom setter methods 
        value.body.append(ast.parse(self._gen_setter_methods))
        
        # @Mstep [] add additional methods to the body
        value.body.append(ast.parse(self._gen_misc_methods))


        value = ast.fix_missing_locations(value)
        return value


    @property
    def result(self)->str:
        '''
            Get this class's source code.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 11:42:40
            `@memberOf`: __init__
            `@property`: result
        '''
        # value = self.result
        return ast.unparse(self.ast)


    @property
    def instantiate_call(self):
        '''
            Get the statement used to instantiate this class.
            
            
            ClassName(some,args)

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 12:04:15
            `@memberOf`: __init__
            `@property`: instantiate_call
        '''
        return ast.unparse(self.instantiate_call_ast)

    # def instantiate_call(self,args:str=None):
    #     value = f"{self.name.name}({args})"
    #     value = ast.parse(value)
    #     return value

    @property
    def instantiate_call_ast(self):
        '''
            Get the instantiate_call_ast property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 12:04:15
            `@memberOf`: __init__
            `@property`: instantiate_call_ast
        '''

        args = ','.join([x.name.name for x in self.arguments])

        value = f"{self.name.name}({args})"
        value = ast.parse(value)
        return value


    # @property
    # def gen_init(self)->config._method_type:
    #     '''
    #         Get the __init__ method for this class or create it.

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-04-2023 14:29:44
    #         `@memberOf`: __init__
    #         `@property`: gen_init
    #     '''

    #     value = self.get_method("__init__")
    #     if value is None:
    #         init = self.add_method("__init__")

    #         initsuper = "super().__init__()" if len(self.bases) else ""

    #         ib = self.init_body
    #         if ib is None:
    #             ib = "pass"
    #         if len(initsuper) > 0:
    #             if "super().__init__" not in ib:
    #                 ib = f"{initsuper}\n{ib}"

    #         value = self.add_method(
    #             "__init__",
    #             description=self.description,
    #             body=ib
    #         )
    #     return value

    # @property
    # def result(self)->str:
    #     '''
    #         Generate the class declaration text.

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-04-2023 09:34:33
    #         `@memberOf`: __init__
    #         `@property`: result
    #     '''
    #     print(f"self.module:{self.module}")
    #     if self.is_dataclass is True:
    #         if self.module is not None:
    #             self.module.add_import("dataclass","dataclasses",None,True)
    #     dataclass_val = "@dataclass" if self.is_dataclass is True else ""
    #     self._apply_init_arg_assignments()
    #     # initsuper = "super().__init__()" if len(self.bases) else ""
    #     s = Template(config.get_template("class_template"))
    #     value = s.substitute(
    #         dataclass=dataclass_val,
    #         class_name=self.name,
    #         bases=self._gen_bases,
    #         dataclass_properties=self._dataclass_properties,
    #         init_method=self.gen_init.result,
    #         # initsuper=initsuper,
    #         getter_methods=self._gen_getter_methods,
    #         setter_methods=self._gen_setter_methods,
    #         misc_methods=self._gen_misc_methods,
    #         timestamp=datetime.datetime.today().strftime("%m-%d-%Y %H:%M:%S"),
    #     )
    #     return value








# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import




import os
import re
import datetime
from dataclasses import dataclass
from typing import Iterable
import colemen_utils as c
from string import Template

import config
log = c.con.log


SINGLE_ARG_TEMPLATE='''    $name {$data_type}
        $description'''

DOCBLOCK_TEMPLATE = '''    $description

    $arguments

    $kwargs

    $returns

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: $timestamp
    `version`: 1.0
    `method_name`: $subject_name
    * @xxx [$timestamp]: documentation for $subject_name'''

@dataclass
class ClassMethodDocBlock:

    pyclass:config._py_class_type = None
    '''A reference to the class that the method belongs to.'''
    class_prop:config._py_property_type = None
    '''A reference to the class property that the method manages. (if it is a getter or setter)'''
    
    return_type:str = None
    '''The datatype returned by this'''
    subject_name:str = None
    '''The name of the subject of this docblock'''
    description:str = None
    '''The description of the docblock'''
    return_description:str = None
    return_type:str = None
    indent:int = 4
    _member_of:str = None

    _arguments=None
    '''A list of argument strings'''
    _kwargs=None


    def __init__(self,classMethod:config._method_type,data:dict = None) -> None:
        self._arguments = []
        self._kwargs = []
        self.classMethod = classMethod
        if data is not None:
            if isinstance(data,(dict)):
                for k,v in data.items():
                    if hasattr(self,k):
                        setattr(self,k,v)

        self.pyclass = self.classMethod.pyclass
        # if hasattr(self.classMethod,'name'):
        self.subject_name = self.classMethod.name.name
        if self.classMethod.name.name == "__init__":
            self.subject_name = self.classMethod.pyclass.name.name
            self.description = self.classMethod.pyclass.description

        if hasattr(self.classMethod,'description'):
            self.description = self.classMethod.description

        self.return_type = self.classMethod.return_type
        self.return_description = self.classMethod.return_description

        if self.classMethod.is_setter or self.classMethod.is_getter:
            self.class_prop = self.pyclass.get_property(self.subject_name)
            if self.class_prop is not None:
                self.return_description = self.class_prop.description







    def add_argument(self,name,data_type,description,default="__NO_DEFAULT_PROVIDED__"):
        s = Template(SINGLE_ARG_TEMPLATE)

        name = f"`{name}`"
        if data_type is None:
            data_type = "any"
        if default != "__NO_DEFAULT_PROVIDED__":
            if isinstance(default,(str)):
                default = f"'{default}'"
            name = f"[{name}={default}]"

        value = s.substitute(
            name=name,
            data_type=data_type,
            description=description,
        )
        self._arguments.append(value)


    def add_kwarg(self,name,data_type,description,default="__NO_DEFAULT_PROVIDED__"):
        s = Template(SINGLE_ARG_TEMPLATE)

        name = f"`{name}`"
        if data_type is None:
            data_type = "any"

        if default != "__NO_DEFAULT_PROVIDED__":
            if isinstance(default,(str)):
                default = f"'{default}'"
            name = f"[{name}={default}]"

        value = s.substitute(
            name=name,
            data_type=data_type,
            description=description,
        )
        self._kwargs.append(value)

    @property
    def arguments(self):
        '''
            Get the arguments property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-15-2022 10:55:22
            `@memberOf`: __init__
            `@property`: arguments
        '''
        self._arguments = []
        value = ''
        for arg in self.classMethod.arg_list:
            if arg.name.name.startswith("**"):
                continue
            self.add_argument(arg.name.name,arg.data_type,arg.description,arg.default)
            args = '\n\n'.join(self._arguments)
            value = f'''Arguments\n    -------------------------\n{args}'''
        return value

    @property
    def kwargs(self):
        '''
            Get the kwargs property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-15-2022 10:55:22
            `@memberOf`: __init__
            `@property`: kwargs
        '''
        self._kwargs = []
        value = ''
        for arg in self.classMethod.kwarg_list:
            self.add_kwarg(arg.name.name,arg.data_type,arg.description,arg.default)
            args = '\n\n'.join(self._kwargs)
            value = f'''Keyword Arguments\n    -------------------------\n{args}'''
        return value


    @property
    def member_of(self)->str:
        '''
            Get the member_of value.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-15-2022 11:04:10
            `@memberOf`: PostArg
            `@property`: member_of
        '''
        if self._member_of is None:
            return ""
        value = f"`memberOf`: {self._member_of}"
        return value

    @member_of.setter
    def member_of(self,value:str):
        '''
            Set the member_of value.

            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-15-2022 11:04:10
            `@memberOf`: PostArg
            `@property`: member_of
        '''
        self._member_of = value


    @property
    def returns(self):
        '''
            Get the returns property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-15-2022 11:00:08
            `@memberOf`: __init__
            `@property`: returns
        '''
        if self.return_type is None and self.return_description is None:
            return ""
        value = f'''Return {{{self.return_type}}}
    ----------------------
    {self.return_description}
'''
        # value = self.returns
        return value


    @property
    def result(self):
        '''
            Get the result property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-08-2022 12:22:06
            `@memberOf`: __init__
            `@property`: result
        '''

        s = Template(DOCBLOCK_TEMPLATE)
        value = s.substitute(
            description=self.description,
            subject_name=self.subject_name,
            member_of=self.member_of,
            timestamp=datetime.datetime.today().strftime("%m-%d-%Y %H:%M:%S"),
            arguments=self.arguments,
            kwargs=self.kwargs,
            returns=self.returns,
        )
        vl = value.split("\n")
        indented = []
        # vl2 = []
        instring = " "*self.indent
        for l in vl:
            if len(c.string.strip(l,[" "])) == 0:
                indented.append("\n")
            else:
                indent = f"\n{instring}"
                indented.append(f"{indent}{l}")
        # vl = vl2
        value = "".join(indented)
        value = re.sub(r"[\s\n]*Meta","\n\n            Meta",value)
        value = re.sub(r"[\s\n]*Keyword",":\n\n            Keyword",value)
        value = re.sub(r"[\s\n]*Return",":\n\n            Return",value)
        if self.classMethod.is_class_method is False:
            value = value.replace("            Return","        Return")
            value = value.replace("            Meta","        Meta")
        # value = re.sub(r":[\s\n]*'''",":\n        '''",value)
        return value


# if __name__ == '__main__':
#     d = DocBlock()
#     d.description = "Some Description"
#     d.subject_name = "boobs"
#     d.add_argument("titties","nips","The titties and the nips")
#     d.add_argument("boobies","butts","The boobies and the butts","pussyFist")
#     # d.member_of = "ass"
#     d.return_type = "str"
#     d.return_description = "That stuff from the thing"
#     print(f"d.result")
#     print(d.result)








# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import




import os
import re
import datetime
from dataclasses import dataclass
from typing import Iterable
import silent.EntityBase as _eb
import colemen_utils as c
from string import Template

import config
log = c.con.log


DOCBLOCK_TEMPLATE='''    $description

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: $timestamp
    `version`: 1.0
    `package_name`: $subject_name
    `member_of`: $package_name
    * @xxx [$timestamp]: documentation for $subject_name
'''


@dataclass
class PackageDocBlock(_eb.EntityBase):
    indent:int = 0

    def __init__(self,main:config._main_type,package:config._package_type,**kwargs):
        '''
            Create a new package docblock.

            ----------

            Arguments
            -------------------------
            `main` {_main_type}
                A reference to the project instance

            `package` {Package}
                A reference to the Package instance

            Keyword Arguments
            -------------------------
            [`name`=None] {str}
                The name of the package.
                This will default to the package's actual name.

            [`description`=None] {str}
                The description of the package
                This will default to the package's description.

            [`indent`=0] {int}
                The indentation to apply the docblock

            Return {type}
            ----------------------
            return_description

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-06-2023 11:10:11
            `memberOf`: PackageDocBlock
            `version`: 1.0
            `method_name`: PackageDocBlock
            * @xxx [01-06-2023 11:21:10]: documentation for PackageDocBlock
        '''
        kwargs['main'] = main
        kwargs['package'] = package
        super().__init__(**kwargs)

        if self.name is None:
            self.name = self.package.name.name
        if self.description is None:
            self.description = self.package.description


    @property
    def result(self):
        '''
            Generate the documentation block string.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-08-2022 12:22:06
            `@memberOf`: __init__
            `@property`: result
        '''

        s = Template(DOCBLOCK_TEMPLATE)
        value = s.substitute(
            description=self.description,
            subject_name=self.name.name,
            package_name=self.package.name.name,
            timestamp=datetime.datetime.today().strftime("%m-%d-%Y %H:%M:%S"),
        )
        vl = value.split("\n")
        indented = []
        instring = " "*self.indent
        for l in vl:
            if len(c.string.strip(l,[" "])) == 0:
                indented.append("\n")
            else:
                indent = f"\n{instring}"
                indented.append(f"{indent}{l}")
        value = "".join(indented)
        value = f"'''{value}'''"
        return value




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import




from dataclasses import dataclass
import re
from string import Template
from typing import Iterable, Union
import ast

import colemen_utils as c
import silent.EntityBase as _eb


# import silent.Class.Method as _method
# from config import column_type,_relationship_type,endpointdoc_type,route_type,root_directory,susurrus_type
import config
log = c.con.log


@dataclass
class Property(_eb.EntityBase):


    data_type:str = None
    '''The python data type that this property stores.'''

    default = None
    '''The default value to assign to this property'''


    is_private:bool = False
    '''True if this property is private to its class.'''

    getter_body:str = None
    '''The body of the getter method.'''

    indent:int = 4


    def __init__(self,main:config._main_type,pyClass:config._py_class_type,**kwargs) -> None:
        '''
            Represents a class property.

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-26-2022 08:35:16
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: Table
            * @xxx [12-26-2022 08:36:08]: documentation for Table
        '''
        kwargs['main'] = main
        kwargs['pyclass'] = pyClass
        kwargs['module'] = pyClass.module

        super().__init__(**kwargs)

        if isinstance(self.default,(str)) and isinstance(self.data_type,(str)):
            if self.data_type in ["int","integer"]:
                self.data_type = "int"
                if c.valid.numeric_only(self.default):
                    self.default = int(self.default)









    @property
    def summary(self):
        '''
            Get the summary property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-06-2022 12:10:00
            `@memberOf`: __init__
            `@property`: summary
        '''
        value = {
            "name":self.name.name,
            "description":self.description,
            "data_type":self.data_type,
            "default":self.default,
            "is_private":self.is_private,
            # "schema":self.table.database.database,
        }

        return value



    def _gen_getter(self):
        pass



    @property
    def attribute_name(self):
        '''
            Get this Property's attribute_name

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 09:15:52
            `@memberOf`: Property
            `@property`: attribute_name
        '''
        value = self.name.name
        if self.is_private:
            value = f"_{self.name.name}"
        return value

    @property
    def attribute_type(self)->str:
        '''
            Get the attribute_type value.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-26-2022 12:10:46
            `@memberOf`: PostArg
            `@property`: attribute_type
        '''
        value = ""

        if isinstance(self.data_type,(str)):
            value = c.string.strip(self.data_type,":")
            value = f":{value}"
        return value

    @property
    def attribute_description(self):
        '''
            Get this Property's attribute_description

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 11:11:47
            `@memberOf`: Property
            `@property`: attribute_description
        '''
        value = ""
        if isinstance(self.description,(str)):
            if len(self.description) > 0:
                value = c.string.strip(self.description,["'"])
                value = f"'''{value}'''"
        return value

    @property
    def default_value(self)->str:
        '''
            Get the default_value value.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-27-2022 10:25:55
            `@memberOf`: PostArg
            `@property`: default_value
        '''
        value = self.default
        if isinstance(value,(str)):
            value = c.string.strip(value,['"'])
            value = f'"{value}"'
        return value


    @property
    def getter(self):
        '''
            Generate this Property's getter method

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 09:08:27
            `@memberOf`: Property
            `@property`: getter
        '''
        if self.pyclass.is_dataclass and self.is_private is False:
            return ""

        description = self.description
        if description is None or len(description) == 0:
            description = f"Retrieve the {self.name.name} property from {self.pyclass.name}"
        body = self.getter_body
        if body is None or len(body) == 0:
            body = f"        return self.{self.attribute_name}"

        method = self.pyclass.add_method(
            self.name.name,
            description=description,
            body=body,
            return_type=self.data_type,
            is_getter=True
        )
        return method

    # @property
    # def declaration(self):
    #     '''
    #         Get this Property's declaration

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-04-2023 09:17:47
    #         `@memberOf`: Property
    #         `@property`: declaration
    #     '''
    #     # import ast
    #     # ass = ast.AnnAssign(
    #     #     target=[
    #     #         ast.Name(
    #     #             id=self.attribute_name,
    #     #             ctx=ast.Store())
    #     #         ],
    #     #     annotation=ast.Name(id=self.attribute_type,ctx=ast.Load()),
    #     #     value=ast.Constant(
    #     #         value=self.default_value
    #     #     ),
    #     #     simple=1
    #     # )
    #     # value = ast.unparse(ass)
    #     # return " "*self.indent + value
    #     _ = self.getter
    #     s = Template(config.get_template("property_declaration_template"))
    #     value = s.substitute(
    #         attribute_name=self.attribute_name,
    #         attribute_type=self.attribute_type,
    #         default_value=self.default_value,
    #         description=self.attribute_description,
    #     )
    #     return value



    @property
    def declaration_ast(self)->Union[ast.Assign,ast.AnnAssign,ast.Expr]:
        '''
            Get this Property's declaration ast object

            If this property has a description a list will be returned:
            [assignment,description]

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 12:04:27
            `@memberOf`: Property
            `@property`: ast
        '''
        name = self.name.name
        if self.is_private:
            name = f"_{name}"

        if self.data_type is not None:
            value = ast.AnnAssign(
                target=ast.Name(id=name, ctx=ast.Store()),
                annotation=ast.Name(id='str', ctx=ast.Load()),
                value=ast.Constant(value=None),
                simple=1
            )
            if self.data_type is not None:
                value.annotation = ast.Name(id=self.data_type, ctx=ast.Load())
                # Expr(
                #     value=Constant(value='The method name'))
        else:
            value = ast.Assign(
                targets=[
                    ast.Name(id=name, ctx=ast.Store())
                    ],
                value=ast.Constant(value=None)
            )

        if self.default is not None:
            value.value = ast.Constant(value=self.default)

        if self.description is not None:
            description = ast.Expr(value=ast.Constant(value=self.description))
            value = [value,description]

        return value

    @property
    def declaration_result(self):
        '''
            Get this Property's declaration_result

            `default`:None

            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 12:10:47
            `@memberOf`: Property
            `@property`: declaration_result
        '''
        value = self.declaration_ast
        if self.description is not None:
            description = ast.Expr(value=ast.Constant(value=self.description))
            value = [value,description]
        # value = ast.fix_missing_locations(value)
        
        return ast.unparse(value)

    
    def assign_ast(self,value,include_description=False):
        '''
            Get this Property's assign_ast

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 14:00:14
            `@memberOf`: Property
            `@property`: assign_ast
        '''
        assign = f"self.{self.attribute_name}{self.attribute_type} = {value}"
        if include_description is True:
            assign = f"{assign}\n'{self.description}'"
        value = ast.parse(assign)
        return value

    def assign_result(self,value,include_description=False):
        '''
            Get this Property's assign_ast

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 14:00:14
            `@memberOf`: Property
            `@property`: assign_ast
        '''
        value = ast.unparse(self.assign_ast(value,include_description))
        return value



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import




from dataclasses import dataclass
import re
from string import Template
from typing import Iterable
import datetime
import ast

import colemen_utils as c
import silent.EntityBase as _eb

import silent.DocBlock.ClassMethodDocBlock as _docblock
import silent.Method.MethodArgument as _ma
# from config import column_type,_relationship_type,endpointdoc_type,route_type,root_directory,susurrus_type
import config
log = c.con.log


@dataclass
class Method(_eb.EntityBase):



    body:str = None
    '''The method's body'''

    return_description:str = None
    '''A description of the return value'''

    return_type:str = None
    '''The type returned by this method'''

    _is_getter:bool = False
    '''If True the property decorator is added to the method'''

    _is_setter:bool = False
    '''If True the property setter decorator is added to the method'''

    indent:int = 0

    is_class_method:bool = False

    _args:Iterable[config._method_argument_type] = None
    _kwargs:Iterable[config._method_argument_type] = None


    def __init__(self,main:config._main_type,**kwargs) -> None:
        '''
            Class used to generate a python class method.
            ----------

            Arguments
            -------------------------
            `main` {_main_type}
                A reference to the master



            Keyword Arguments
            -------------------------
            `name` {str}
                The name of this method.

            [`pyclass`=None] {Class}
                A reference to the class that this method belongs to.

            [`module`=None] {Module}
                A reference to the module that this method belongs to.

            [`description`=None] {str}
                The docblock description for this method.

            [`body`="pass"] {str}
                The body of the method.

            [`return_type`=None] {str}
                The type returned by this method

            [`return_description`=None] {str}
                A description of the return value

            [`is_getter`=False] {bool}
                If True the property decorator is added to the method

            [`is_setter`=False] {bool}
                If True the property setter decorator is added to the method

            [`is_class_method`=False] {bool}
                True if this method belongs to a class.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-26-2022 08:35:16
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: Table
            * @xxx [12-26-2022 08:36:08]: documentation for Table
        '''
        kwargs['main'] = main
        kwargs['body'] = c.obj.get_kwarg(['body'],'pass',None,**kwargs)
        super().__init__(**kwargs)
        # self.main:config._main_type = main
        # self.pyclass:config._py_class_type = pyclass
        self.Doc = None
        self._args = []
        self._kwargs = []
        # self.Doc = _docblock.ClassMethodDocBlock(self)


        # populate_from_dict(kwargs,self)

        if self.is_class_method is True:
            self.indent = 4
            self.add_tag(self.pyclass.name.name)
        # self.Doc.indent = self.indent + 4

    @property
    def summary(self):
        '''
            Get the summary property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-06-2022 12:10:00
            `@memberOf`: __init__
            `@property`: summary
        '''
        value = {
            "name":self.name.name,
            "description":self.description,
            "is_class_method":self.is_class_method,
            "return_description":self.return_description,
            "return_type":self.return_type,
            "is_getter":self.is_getter,
            "is_setter":self.is_setter,
            "arguments":[],
            "keyword_arguments":[],
            "import_statement":self.import_statement,
        }
        for arg in self.arg_list:
            value['arguments'].append(arg.summary)


        return value


    @property
    def local_name(self)->str:
        '''
            Get this Method's local_name

            The local name is used by the class for distinguishing methods with the same name.
            Specifically, this solves the name collisions caused by getters and setters having the same
            name. It does this by applying a suffix:

            propname____get
            propname____set

            If this method is not a getter or setter, its normal name is returned.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 09:00:57
            `@memberOf`: Method
            `@property`: get_local_name
        '''
        if self.is_getter:
            return f"{self.name.name}____get"
        if self.is_setter:
            return f"{self.name.name}____set"
        return self.name.name


    @property
    def is_getter(self)->bool:
        '''
            Get the is_getter value.

            `default`:False


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 10:09:11
            `@memberOf`: PostArg
            `@property`: is_getter
        '''
        value = self._is_getter
        return value

    @is_getter.setter
    def is_getter(self,value:bool):
        '''
            Set the is_getter value.

            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 10:09:11
            `@memberOf`: PostArg
            `@property`: is_getter
        '''
        if value is True:
            self.add_tag("getter","setter")
        else:
            self.delete_tag("getter")
        self._is_getter = value

    @property
    def is_setter(self)->bool:
        '''
            Get the is_setter value.

            `default`:False


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 10:16:05
            `@memberOf`: PostArg
            `@property`: is_setter
        '''
        value = self._is_setter
        return value

    @is_setter.setter
    def is_setter(self,value:bool):
        '''
            Set the is_setter value.

            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 10:16:05
            `@memberOf`: PostArg
            `@property`: is_setter
        '''
        if value is True:
            self.add_tag("setter","getter")
        else:
            self.delete_tag("setter")
        self._is_setter = value


    # ---------------------------------------------------------------------------- #
    #                                   ARGUMENTS                                  #
    # ---------------------------------------------------------------------------- #



    def add_arg(self,name,**kwargs):
        '''
            Add an argument to this method.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the argument.

            Keyword Arguments
            -------------------------
            [`data_type`=None] {str}
                The expected type of the argument

            [`default`=None] {any}
                The default value to use if no value is provided.

            [`description`=None] {any}
                The documentation description

            [`arg`=None] {Method_Argument}
                A method argument instance to copy.
                This is useful when you want to add the same argument to another method.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-03-2023 15:41:59
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_arg
            * @xxx [01-03-2023 15:43:13]: documentation for add_arg
        '''

        # arg = self.get_arg(name)
        # if arg is not None:
        #     print(f"ARGUMENT ALREADY EXISTS")
        #     return
        # import silent.Method.MethodArgument as _ma
        ma = _ma.MethodArgument(
            self.main,
            self,
            name=name,
            arg = c.obj.get_kwarg(['arg'],None,None,**kwargs),
            data_type = c.obj.get_kwarg(['data_type','type'],None,(str),**kwargs),
            default = c.obj.get_kwarg(['default_value','default'],"__NO_DEFAULT_PROVIDED__",None,**kwargs),
            description = c.obj.get_kwarg(['description'],"",None,**kwargs),
        )
        # data_type = c.obj.get_kwarg(['py_type','type'],None,(str),**kwargs)
        # default_value = c.obj.get_kwarg(['default_value','default'],"__NO_DEFAULT_VALUE_SET__",None,**kwargs)
        # description = c.obj.get_kwarg(['description'],"",None,**kwargs)
        # ma.name = name
        # ma.py_type = data_type
        # ma.default_value = default_value
        # ma.description = description

        if ma.has_default:
            self._args.append(ma)
        else:
            self._args.insert(0,ma)
        # self._args.append(ma)


    def add_kwarg(self,name,**kwargs):
        '''
            Add a keyword argument to this method.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the argument.

            Keyword Arguments
            -------------------------
            [`data_type`=None] {str}
                The expected type of the argument
            [`default`=None] {any}
                The default value to use if no value is provided.
            [`description`=None] {any}
                The documentation description

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-03-2023 15:41:59
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_arg
            * @xxx [01-03-2023 15:43:13]: documentation for add_arg
        '''


        # ma = _ma.MethodArgument(self.main,self)
        ma = _ma.MethodArgument(
            self.main,
            self,
            name=name,
            data_type = c.obj.get_kwarg(['data_type','type'],None,(str),**kwargs),
            default = c.obj.get_kwarg(['default_value','default'],"__NO_DEFAULT_PROVIDED__",None,**kwargs),
            description = c.obj.get_kwarg(['description'],"",None,**kwargs),
        )

        # data_type = c.obj.get_kwarg(['py_type','type'],None,(str),**kwargs)
        # default_value = c.obj.get_kwarg(['default_value','default'],"__NO_DEFAULT_VALUE_SET__",None,**kwargs)
        # description = c.obj.get_kwarg(['description'],"",None,**kwargs)
        # ma.name = name
        # ma.py_type = data_type
        # ma.default_value = default_value
        # ma.description = description

        if ma.has_default:
            self._kwargs.append(ma)
        else:
            self._kwargs.insert(0,ma)
        # self._args.append(ma)



    def get_arg(self,name:str):
        for arg in self.arg_list:
            if arg.name.name == name:
                return arg
        return None

    # @property
    # def _gen_args_result(self)->str:
    #     '''
    #         Generate the arguments list for this method.

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-03-2023 12:21:19
    #         `@memberOf`: __init__
    #         `@property`: args
    #     '''
    #     value = ""

    #     if len(self._args) > 0:
    #         args = []
    #         for arg in self._args:
    #             args.append(arg.result)
    #         alist = ','.join(args)
    #         value = f",{alist}"
    #     return value

    @property
    def arg_list(self)->Iterable[config._method_argument_type]:
        '''
            Get a list of method argument instances

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-03-2023 16:00:40
            `@memberOf`: __init__
            `@property`: arg_list
        '''
        value = self._args
        return value

    @property
    def kwarg_list(self)->Iterable[config._method_argument_type]:
        '''
            Get the kwarg_list property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-03-2023 16:23:06
            `@memberOf`: __init__
            `@property`: kwarg_list
        '''
        value = self._kwargs
        return value









    @property
    def _return_type(self):
        '''
            Get the _return_type property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-03-2023 12:02:39
            `@memberOf`: __init__
            `@property`: _return_type
        '''
        value = self.return_type
        if value is None:
            return ""
        if isinstance(value,(str)):
            value = c.string.strip(value,['-','>'])
            value = f"->{value}"
        return value

    @property
    def decorator(self):
        '''
            Get the decorator property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-03-2023 12:05:57
            `@memberOf`: __init__
            `@property`: decorator
        '''
        value = ""
        if self.is_getter is True:
            value = "@property"
        if self.is_setter is True:
            value = f"@{self.name}.setter"
        return value

    @property
    def _body(self):
        '''
            Get this Method's _body

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 14:14:46
            `@memberOf`: Method
            `@property`: _body
        '''
        value = self.body
        if value is None:
            return "pass"
        return value
            # return " "*(self.indent+4) + "pass"
        values = value.split("\n")
        # value = '\n'.join([f"{' '*(self.indent+4)}{x}" for x in values])
        value = '\n'.join([x for x in values])

        return value

    # @property
    # def result(self):
    #     '''
    #         Get the result property's value

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 12-28-2022 16:49:44
    #         `@memberOf`: __init__
    #         `@property`: result
    #     '''
    #     self.Doc.subject_name = self.name
    #     self.Doc.description = self.description
    #     s = Template(config.get_template("method_template"))
    #     value = s.substitute(
    #         decorator=self.decorator,
    #         timestamp=datetime.datetime.today().strftime("%m-%d-%Y %H:%M:%S"),
    #         method_name=self.name,
    #         return_type=self._return_type,
    #         body=self._body,
    #         arguments=self._gen_args_result,
    #         description=self.description,
    #         docblock=self.Doc.result,
    #     )
    #     value = re.sub(r"[\s\n]*Meta","\n\n            Meta",value)
    #     value = re.sub(r":[\s\n]*'''",":\n        '''",value)
    #     return value


    @property
    def _gen_return_type(self):
        '''
            Get this Method's _gen_return_type

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 15:54:20
            `@memberOf`: Method
            `@property`: _gen_return_type
        '''
        if self.return_type is None:
            return ast.Constant(value=None)


        value = ast.Attribute(ctx=ast.Load())
        types = self.return_type.split(".")
        if len(types) == 1:
            # value.value = ast.Name(id=types[0], ctx=ast.Load())
            return ast.Name(id=types[0], ctx=ast.Load())

        if len(types) == 2:
            value.value=ast.Attribute(
                value=ast.Name(id=types[0], ctx=ast.Load()),
                attr=types[1],
                ctx=ast.Load()
            )
        if len(types) > 2:
            ann = ast.Attribute(
                value=ast.Name(id=types[0], ctx=ast.Load()),
                attr=types[1],
                ctx=ast.Load()
            )
            # @Mstep [] remove the first and second element.
            types = types[2:]

            for typ in types:
                ann=ast.Attribute(
                    value=ann,
                    attr=typ,
                    ctx=ast.Load()
                )

            value.value = ann
        return value

    @property
    def declaration_ast(self):
        '''
            Get this Method's ast

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 12:34:44
            `@memberOf`: Method
            `@property`: ast
        '''

        # return_type = ast.parse(self.return_type) if self.return_type is not None else ast.Constant(value=None)
        value = ast.FunctionDef(
            name=self.name.name,
            args=ast.arguments(
                posonlyargs=[],
                args=[],
                kwonlyargs=[],
                kw_defaults=[],
                # TODO []: add **kwargs to method.
                kwarg=[],
                defaults=[]),
            body=[],
            decorator_list=[],
            returns=self._gen_return_type
        )
        # if self.return_type is not None:
        #     print(f"return_type: {self.return_type}")
        #     value.returns = ast.parse(self.return_type)
            # print(ast.dump(returns))

        if self.is_class_method:
            # @Mstep [] add the "self" argument to the method.
            value.args.args.insert(0,ast.arg(arg='self'))
            # @Mstep [] add the getter/setter decorators.
            if self.is_getter:
                value.decorator_list.append(ast.Name(
                    id='property',
                    ctx=ast.Load()
                    )
                )
            if self.is_setter:
                attr = ast.Attribute(
                            value=ast.Name(
                                id=self.name.name,
                                ctx=ast.Load()
                            ),
                            attr='setter',
                            ctx=ast.Load()
                        )
                value.decorator_list.append(attr)

        # print(f"self.arg_list: {len(self.arg_list)}")
        for arg in self.arg_list:
            value.args.args.append(arg.ast)

        # @Mstep [IF] if there are keyword arguments
        if len(self.kwarg_list) > 0:
            # @Mstep [] add the keyword argument variable to the method.
            value.args.kwarg = ast.arg(arg='kwargs')

        # @Mstep [] apply the body.
        value.body.append(ast.parse(self._body))

        # @Mstep [] prepend the docblock to the method body.
        # self.Doc.subject_name = self.name.name
        # self.Doc.description = self.description
        self.Doc = _docblock.ClassMethodDocBlock(self)
        self.Doc.indent = self.indent + 4
        value.body.insert(0,ast.Expr(value=ast.Constant(value=self.Doc.result)))



        value = ast.fix_missing_locations(value)
        return value


# def populate_from_dict(data:dict,instance:Method):
#     for k,v in data.items():
#         if hasattr(instance,k):
#             setattr(instance,k,v)





from setuptools import setup, find_packages
import build_utils as _bu
PY_MODULES = _bu.list_py_modules()

VERSION = '0.0.0'
DESCRIPTION = 'colemen_silentEngine'
LONG_DESCRIPTION = 'colemen_silentEngine'

# Setting up
setup(
    name="colemen_silent_engine",
    version=VERSION,
    author="Colemen Atwood",
    author_email="<atwoodcolemen@gmail.com>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    py_modules=['main'] + PY_MODULES,
    # add any additional packages that
    # need to be installed along with your package. Eg: 'caer'
    install_requires=[
        'colemen_utils',
    ],

    keywords=['python'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import




from dataclasses import dataclass
from string import Template
from typing import Iterable
import datetime
import ast

import colemen_utils as c
import silent.EntityBase as _eb

# import agod.TypeDeclaration as _type_dec
# from config import column_type,_relationship_type,endpointdoc_type,route_type,root_directory,susurrus_type
import config
log = c.con.log


@dataclass
class MethodArgument(_eb.EntityBase):


    arg:config._method_argument_type = None
    '''A reference to another method argument to copy data from.'''
    
    data_type:str = None
    '''The expected python type for this argument'''

    default = "__NO_DEFAULT_VALUE_SET__"
    '''The default value to assign if the argument is not provided.'''




    def __init__(self,main:config._main_type,method:config._method_type,**kwargs) -> None:
        '''
            A class used to represent an argument passed to a Method.
            
            ----------


            Arguments
            -------------------------
            `main` {_main_type}
                A reference to the master

            `method` {Method}
                A reference to the Method this argument belongs to.

            [`pyclass`=None] {Class}
                A reference to the class that this method belongs to.


            Keyword Arguments
            -------------------------
            `name` {str}
                The name of this method.

            [`description`=None] {str}
                The docblock description for this method.

            [`default`=None] {str}
                The default value for this argument

            [`arg`=None] {Method_Argument}
                A method argument instance to copy.
                This is useful when you want to add the same argument to another method.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-26-2022 08:35:16
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: Table
            * @xxx [12-26-2022 08:36:08]: documentation for Table
        '''
        kwargs['main'] = main
        kwargs['method'] = method
        super().__init__(**kwargs)
        # self.main:config._main_type = main

        # self.class_method:config._method_type = class_method

    @property
    def summary(self):
        '''
            Get the summary property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-06-2022 12:10:00
            `@memberOf`: __init__
            `@property`: summary
        '''
        value = {
            "name":self.name.name,
            "description":self.description,
            "default":self.default,
        }
        return value

    @property
    def has_default(self):
        '''
            Get this MethodArgument's has_default

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-03-2023 12:22:13
            `@memberOf`: MethodArgument
            `@property`: has_default
        '''
        value = False
        if self.default != "__NO_DEFAULT_VALUE_SET__":
            value = True
        return value

    # @property
    # def _data_type(self):
    #     '''
    #         Get this MethodArgument's _data_type

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-03-2023 12:09:51
    #         `@memberOf`: MethodArgument
    #         `@property`: _data_type
    #     '''
    #     value =""
    #     if self.data_type is not None:
    #         value = f":{self.data_type}"
    #     return value

    # @property
    # def _default_value(self):
    #     '''
    #         Get this MethodArgument's _default_value

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-03-2023 12:11:30
    #         `@memberOf`: MethodArgument
    #         `@property`: _default_value
    #     '''
    #     value = ""
    #     if self.default != "__NO_DEFAULT_VALUE_SET__":
    #         value = f"={self.default}"
    #     return value

    @property
    def name_type(self):
        '''
            Get this MethodArgument's name and type as a string
            
            arg:str

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 12:09:57
            `@memberOf`: MethodArgument
            `@property`: name_type
        '''
        value = f"{self.name.name}"
        if self.data_type is not None:
            value = f"{value}:{self.data_type}"
        if self.default != "__NO_DEFAULT_PROVIDED__":
            value = f"{value}={self.default}"
        return value

    @property
    def ast(self)->ast.arg:
        '''
            Get this MethodArgument's ast

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 12:46:57
            `@memberOf`: MethodArgument
            `@property`: ast
        '''
        
        if self.arg is not None:
            # import silent.Method.MethodArgument as _ma
            if isinstance(self.arg,MethodArgument):
                arg = self.arg
                self.name = arg.name.name
                self.description = arg.description
                self.data_type = arg.data_type
        
        value = ast.arg(
            arg=self.name.name,
        )
        if self.data_type is not None:
            types = self.data_type.split(".")
            if len(types) == 1:
                value.annotation = ast.Name(id=types[0], ctx=ast.Load())

            if len(types) == 2:
                value.annotation=ast.Attribute(
                    value=ast.Name(id=types[0], ctx=ast.Load()),
                    attr=types[1],
                    ctx=ast.Load()
                )
            if len(types) > 2:
                ann = ast.Attribute(
                    value=ast.Name(id=types[0], ctx=ast.Load()),
                    attr=types[1],
                    ctx=ast.Load()
                )
                # @Mstep [] remove the first and second element.
                types = types[2:]
                # @Mstep [] reverse the list.
                # types.reverse()

                # if isinstance(types,(list)):
                # atypes = []
                for typ in types:
                    ann=ast.Attribute(
                        value=ann,
                        attr=typ,
                        ctx=ast.Load()
                    )

                value.annotation = ann

        return value

    # @property
    # def result(self):
    #     '''
    #         Get the result property's value

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 12-28-2022 16:49:44
    #         `@memberOf`: __init__
    #         `@property`: result
    #     '''
    #     return f"{self.name}{self._data_type}{self._default_value}"

